#include "world_view.h"
#include "l3d.h"
#include "frame.h"
#include "player.h"
#include "space.h"

static const float lightCol[] = { 1, 1, .9, 0 };

#define BG_STAR_MAX 2000

WorldView::WorldView(void): View() {
  SetTransparency(true);
  
  m_hyperspaceButton = new Gui::ImageButton("icons/hyperspace_f8.png");
  m_hyperspaceButton->SetShortcut(SDLK_F8, KMOD_NONE);
  m_hyperspaceButton->onClick.connect(sigc::mem_fun(this, &WorldView::OnClickHyperspace));
  m_rightButtonBar->Add(m_hyperspaceButton, 66, 2);

  m_bgstarsDlist = glGenLists(1);

  glNewList(m_bgstarsDlist, GL_COMPILE);

  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);
  glPointSize(1.0);
  glBegin(GL_POINTS);
  for(int i = 0; i < BG_STAR_MAX; i++) {
    float col = 0.2+L3D::rng(0.8);
    glColor3f(col, col, col);
    glVertex3f(1000-L3D::rng(2000.0), 1000-L3D::rng(2000.0), 1000-L3D::rng(2000.0));
  }

  glEnd();
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);

  glEndList();
}

void WorldView::OnClickHyperspace(void) {
  StarSystem* s = L3D::GetSelectedSystem();
  if(s /* && isn's current system. */) {
    printf("Traveling through hyperspace. WWEEEEEE!!\n");
    L3D::HyperspaceTo(s);
  }
}

void WorldView::Draw3D(void) {
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  /* Wth did I give these functions large names for.. :/ */
  glFrustum(-L3D::GetScrWidth()*.5, L3D::GetScrWidth()*.5,
            -L3D::GetScrHeight()*.5, L3D::GetScrHeight()*.5,
             L3D::GetScrWidth()*.5, 100000);
  glDepthRange(-10, -100000);
  glMatrixMode(GL_MODELVIEW);

  /* Make temporary camera frame at player. */
  Frame* cam_frame = new Frame(L3D::player->GetFrame(), "", Frame::TEMP_VIEWING);

  if(L3D::GetCamType() == L3D::CAM_FRONT) {
    cam_frame->SetPosition(L3D::player->GetPosition());
  } else if(L3D::GetCamType() == L3D::CAM_REAR) {
    glRotatef(180.0f, 0, 1, 0);
    cam_frame->SetPosition(L3D::player->GetPosition());
  } else { /* CAM_EXTERNAL */
    cam_frame->SetPosition(L3D::player->GetPosition());
    L3D::player->ApplyExternalViewRotation();
  }
    
  L3D::player->ViewingRotation();

  glGetDoublev(GL_MODELVIEW_MATRIX, &viewingRotation[0]);

  glCallList(m_bgstarsDlist);
  /* Position light at sol. */
  vector3d lpos = Frame::GetFramePosRelativeToOther(Space::GetRootFrame(), cam_frame);
  float lightPos[4];
  lightPos[0] = lpos.x;
  lightPos[1] = lpos.y;
  lightPos[2] = lpos.z;
  lightPos[3] = 0;
  glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
  glLightfv(GL_LIGHT0, GL_AMBIENT_AND_DIFFUSE, lightCol);
  glLightfv(GL_LIGHT0, GL_SPECULAR, lightCol);

  Space::Render(cam_frame);
  L3D::player->DrawHUD(cam_frame);
  
  L3D::player->GetFrame()->RemoveChild(cam_frame);
  delete cam_frame;
}

void WorldView::Update(void) {
  if(L3D::GetSelectedSystem() /* && isn't current system */) {
    m_hyperspaceButton->Show();
  } else {
    m_hyperspaceButton->Hide();
  }
  /* Player control inputs. */
  L3D::player->AITurn();
}

